
#ifndef _EDITOR_BASE_H_
#define _EDITOR_BASE_H_


#include "ui_form.h"
#include <grtpp.h>
#include "grt/grt_manager.h"
#include <grtpp_undo_manager.h>

#include "grt_backend_public_interface.h"


// How Editor Undo Works
//
// - Each editor has an undo stack
// - When the editor is created, it will save the ref to the object being
// edited and take a snapshot of it
// - Any changes are made to the snapshot copy
// - Whenever a change is started, a snapshot of the snapshot is taken
// and when the change is finished, the diff between snapshots is saved in the stack
// and the new snapshot replaces the old one, which is discarded
// - When an undo is requested, a snapshot is also created and diffed, for use in redo
// - When the user presses Apply, the latest version of object is diffed against the
// live object and the changes are applied to the live object
//
class GrtObject;


namespace bec {
  struct ChangeSet;

  struct MYSQLGRTBACKEND_PUBLIC_FUNC RefreshCentry
  {
    BaseEditor &_editor;
    RefreshCentry(BaseEditor &editor);

    ~RefreshCentry();
  };


  class MYSQLGRTBACKEND_PUBLIC_FUNC BaseEditor : public UIForm 
  {
  public:
    BaseEditor(GRTManager *grtm, const grt::Ref<GrtObject> &object);
    virtual ~BaseEditor();

    virtual bool should_close_on_delete_of(const std::string &oid) { return get_object().id() == oid; }

    // must return a copy of the object
    virtual grt::Ref<GrtObject> get_object()= 0;

    //sigc::signal<void, BaseEditor*> signal_unlocked() { return _unlocked_signal; };
    void add_unlocked_handler(const sigc::slot<void, BaseEditor*>& slot) { _unlocked_signal.connect(slot); }

    grt::GRT *get_grt() { return _grtm->get_grt(); }
    GRTManager *get_grt_manager() { return _grtm; }

    virtual grt::UndoManager *get_undo_manager() { return get_grt()->get_undo_manager(); }

    bool add_undo_object_text_change(const grt::ObjectRef &object, const std::string &member);
    bool begin_undo_object_text_change(const grt::ObjectRef &object, const std::string &member);
    void add_undo(grt::UndoAction *action, const std::string &description= "");

    void begin_undo_group();
    virtual void end_undo_group(const std::string &description="");
    void cancel_undo_group();

    void set_refresh_ui_slot(const sigc::slot<void> &slot);

  public: // of interest for frontend
    // lock when frontend starts using the editor, unlock when done
    void lock();
    void unlock();

    void block_auto_refresh() { _refresh_ui.block(); }
    void unblock_auto_refresh()  { _refresh_ui.unblock(); }

  protected:
    GRTManager *_grtm;

    std::set<std::string> _ignored_object_fields_for_ui_refresh;
    sigc::slot<void> _refresh_ui;

    sigc::signal<void, BaseEditor*> _unlocked_signal;

    void add_listeners(const grt::Ref<GrtObject> &object);

    void run_from_grt(const sigc::slot<void> &slot);

  private:
    grt::Ref<GrtObject> _object;
    void object_member_changed(const std::string &member, const grt::ValueRef &ovalue);
    virtual void on_object_changed();

    void undo_applied();

    bool _locked;
  };

};

#endif /* _EDITOR_BASE_H_ */
