#include "mdc_gtk_canvas_view.h"
#include <gdk/gdkx.h>
#include <sys/time.h>

#define WHEEL_SCROLL_STEP 12

using namespace mdc;





std::string mdc::detect_opengl_version()
{
  int major, minor;
  
  if (!glXQueryVersion(gdk_display, &major, &minor))
    return "";  // glx not supported

  /*
  char buffer[100];
  
  sprintf(buffer, "Vendor: %s, Renderer: %s, OpenGL version: %s, GLX version: %i.%i",
          glGetString(GL_VENDOR),
          glGetString(GL_RENDERER),
          glGetString(GL_VERSION),
          major, minor);*/

//  return (const char*)glGetString(GL_VERSION) ?:"";

  return "";
}




GtkCanvas::GtkCanvas(CanvasType type)
  : _canvas(0), _canvas_type(type)
{
  _hadjustment= 0;
  _vadjustment= 0;

  set_flags(get_flags()|Gtk::CAN_FOCUS|Gtk::APP_PAINTABLE);
  modify_bg(Gtk::STATE_NORMAL, get_style()->get_white());
  set_double_buffered(false);
  add_events(Gdk::POINTER_MOTION_MASK|Gdk::BUTTON_PRESS_MASK|Gdk::BUTTON_RELEASE_MASK|Gdk::KEY_PRESS_MASK|Gdk::KEY_RELEASE_MASK);
  signal_expose_event().connect(sigc::mem_fun(this, &GtkCanvas::redraw));
}


void GtkCanvas::on_realize()
{ 
  Gtk::DrawingArea::on_realize();

  switch (_canvas_type)
  {
  case OpenGLCanvasType:
    _canvas= new GLXCanvasView(gdk_display, gdk_x11_drawable_get_xid(get_window()->gobj()),
                               gdk_x11_visual_get_xvisual(get_visual()->gobj()),
                               get_width(), get_height());
    break;
  case XlibCanvasType:
    _canvas= new XlibCanvasView(gdk_display, gdk_x11_drawable_get_xid(get_window()->gobj()),
                                gdk_x11_visual_get_xvisual(get_visual()->gobj()), get_width(), get_height());
    break;
  case BufferedXlibCanvasType:
    _canvas= new BufferedXlibCanvasView(gdk_display, gdk_x11_drawable_get_xid(get_window()->gobj()),
                                gdk_x11_visual_get_xvisual(get_visual()->gobj()), gdk_drawable_get_depth(get_window()->gobj()),
                                get_width(), get_height());
    break;
  }
  _initialized= false;
}


void GtkCanvas::on_map()
{
  Gtk::DrawingArea::on_map();
  
  if (_initialized)
    return;
  
  if (!_canvas->initialize())
  {
    g_warning("could not initialize canvas");
    delete _canvas;
    _canvas= 0;
    return;
  }

  _canvas->signal_viewport_changed().connect(sigc::mem_fun(this, &GtkCanvas::canvas_view_viewport_changed));
  _canvas->signal_repaint().connect(sigc::mem_fun(this, &GtkCanvas::canvas_view_needs_repaint));

  _canvas->repaint();
}


bool GtkCanvas::on_configure_event(GdkEventConfigure *event)
{
  Gtk::DrawingArea::on_configure_event(event);

  if (_canvas)
    _canvas->update_view_size(event->width, event->height);

  return true;
}


mdc::EventState GtkCanvas::get_event_state(int event_state)
{
  mdc::EventState state= ((mdc::EventState)0);

  if (event_state & GDK_CONTROL_MASK)
    state= state | mdc::SControlMask;
  if (event_state & GDK_SHIFT_MASK)
    state= state | mdc::SShiftMask;
  if (event_state & GDK_MOD1_MASK)
    state= state | mdc::SAltMask;
  
  if (event_state & GDK_BUTTON1_MASK)
    state= state | mdc::SLeftButtonMask;
  if (event_state & GDK_BUTTON2_MASK)
    state= state | mdc::SMiddleButtonMask;
  if (event_state & GDK_BUTTON3_MASK)
    state= state | mdc::SRightButtonMask;
  
  return state;
}


bool GtkCanvas::on_button_press_event(GdkEventButton* event)
{
  MouseButton button= ButtonLeft;
  
  grab_focus();

  switch (event->button)
  {
  case 1: button= ButtonLeft; break;
  case 2: button= ButtonMiddle; break;
  case 3: button= ButtonRight; break;
  }

  _canvas->handle_mouse_button(button, true,
                               (int)event->x, (int)event->y,
                               get_event_state(event->state));

  return true;
}


bool GtkCanvas::on_button_release_event(GdkEventButton* event)
{
  MouseButton button= ButtonLeft;

  switch (event->button)
  {
  case 1: button= ButtonLeft; break;
  case 2: button= ButtonMiddle; break;
  case 3: button= ButtonRight; break;
  }

  _canvas->handle_mouse_button(button, false,
                               (int)event->x, (int)event->y,
                               get_event_state(event->state));

  return true;
}


bool GtkCanvas::on_motion_notify_event(GdkEventMotion* event)
{
  _canvas->handle_mouse_move((int)event->x, (int)event->y, 
                             get_event_state(event->state));
  
  return true;
}



bool GtkCanvas::on_scroll_event(GdkEventScroll *event)
{
  double x, y;
  mdc::Rect rect;
  rect= _canvas->get_viewport();
  
  x= rect.pos.x;
  y= rect.pos.y;
  
  switch (event->direction)
  {
  case GDK_SCROLL_DOWN:
    if ((event->state & GDK_CONTROL_MASK) == 0)
    {
      y+= WHEEL_SCROLL_STEP;
      break;
    }
  case GDK_SCROLL_RIGHT:
    x+= WHEEL_SCROLL_STEP;
    break;
    
  case GDK_SCROLL_UP:
    if ((event->state & GDK_CONTROL_MASK) == 0)
    {
      y-= WHEEL_SCROLL_STEP;
      break;
    }
  case GDK_SCROLL_LEFT:
    x-= WHEEL_SCROLL_STEP;
    break;
  }
  
  if (_vadjustment)
  {
    if (y < 0) y = 0; else if (y > _vadjustment->get_upper()) y = _vadjustment->get_upper();
    
    if (_vadjustment->get_value() != y)
      _vadjustment->set_value(y);
  }
   
  if (_hadjustment)
  {
    if (x < 0) x = 0; else if (x > _hadjustment->get_upper()) x = _hadjustment->get_upper();
    
    if (_hadjustment->get_value() != x)
      _hadjustment->set_value(x);
  }
  
  return true;
}



bool GtkCanvas::on_key_press_event(GdkEventKey *event)
{
  return false;
}


bool GtkCanvas::on_key_release_event(GdkEventKey *event)
{
  return false;
}


bool GtkCanvas::redraw(GdkEventExpose *ev)
{
//  if (ev->count == 0)
  {
    struct timeval tv, tv2;

//    printf("repaint %i %i %i %i\n", ev->area.x, ev->area.y, ev->area.width, ev->area.height);

    gettimeofday(&tv, NULL);
    _canvas->repaint(ev->area.x, ev->area.y, ev->area.width, ev->area.height);
    gettimeofday(&tv2, NULL);

    static const char *debug_canvas= getenv("DEBUG_CANVAS");
    if (debug_canvas)
      printf("rendertime= %.4f (%.1ffps)\n", (tv2.tv_sec - tv.tv_sec) + (tv2.tv_usec - tv.tv_usec)/1000000.0,
             1.0/((tv2.tv_sec - tv.tv_sec) + (tv2.tv_usec - tv.tv_usec)/1000000.0));
  }

  return true;
}



void GtkCanvas::canvas_view_needs_repaint(int,int,int,int)
{
  queue_draw();
}


void GtkCanvas::canvas_view_viewport_changed()
{
  update_scrollers();
}


void GtkCanvas::set_scroll_adjustments(Gtk::Adjustment* hadjustment,
                                      Gtk::Adjustment* vadjustment)
{
  _hadjustment= hadjustment;
  _vadjustment= vadjustment;
  
  _hadjustment->signal_value_changed().connect(sigc::mem_fun(*this, &GtkCanvas::scroll_canvas));
  _vadjustment->signal_value_changed().connect(sigc::mem_fun(*this, &GtkCanvas::scroll_canvas));
  if (_canvas)
    update_scrollers();
}


void GtkCanvas::scroll_canvas()
{
  if (_canvas)
  {
    float xpos= _hadjustment->get_value();
    float ypos= _vadjustment->get_value();

    _canvas->set_offset(Point(xpos, ypos));
  }
}


void GtkCanvas::update_scrollers()
{
  mdc::Size size= _canvas->get_total_view_size();
  mdc::Rect vp= _canvas->get_viewport();

  if (_hadjustment)
  {
    if (_hadjustment->get_lower() != 0)
      _hadjustment->set_lower(0);
    if (_hadjustment->get_upper() != size.width)
      _hadjustment->set_upper(size.width);
    if (_hadjustment->get_page_increment() != vp.size.width/2)
      _hadjustment->set_page_increment(vp.size.width/2);
    if (_hadjustment->get_page_size() != vp.size.width)
      _hadjustment->set_page_size(vp.size.width);
    if (_hadjustment->get_step_increment() != 10)
      _hadjustment->set_step_increment(10);
    if (_hadjustment->get_value() != vp.pos.x)
      _hadjustment->set_value(vp.pos.x);
  }
  if (_vadjustment)
  {
    if (_vadjustment->get_lower() != 0)
      _vadjustment->set_lower(0);
    if (_vadjustment->get_upper() != size.height)
      _vadjustment->set_upper(size.height);
    if (_vadjustment->get_page_increment() != vp.size.height/2)
      _vadjustment->set_page_increment(vp.size.height/2);
    if (_vadjustment->get_page_size() != vp.size.height)
      _vadjustment->set_page_size(vp.size.height);
    if (_vadjustment->get_step_increment() != 10)
      _vadjustment->set_step_increment(10);
    if (_vadjustment->get_value() != vp.pos.y)
      _vadjustment->set_value(vp.pos.y);
  }
}
